/**********************************************************************
*                                                                     *
*    Filename:      controlio.c                                       *
*    Date:                                                            *
*    File Version:                                                    *
*                                                                     *
*    Author:        Peter Jakab <el@jap.hu>                           *
*                   http://jap.hu/electronic/                         *
*                                                                     *
***********************************************************************
NOTES

USB digital I/O extender interface
uses libusb

***********************************************************************
HISTORY

001-20100919

*/

#include <stdio.h>
#include <usb.h>

#define FALSE   0
#define TRUE    1

#define	SETPORT			0x54
#define	GETPORT			0x53
#define SETDIR			0x55
#define SETBIT			0x56
#define CLEARBIT		0x57

int main(int argc, char *argv[]) {
    struct usb_bus *bus;
    struct usb_device *dev;
    usb_dev_handle *udev;
    int dev_found, ret;
    char buffer[8];
    unsigned int bRequest = 0; // request
    unsigned int wValue = 0; // request data
    unsigned int wIndex = 0; // not used, optional data
    unsigned int wLength = 0; // length of buffer[]
    unsigned int bmRequestType = USB_TYPE_VENDOR | USB_RECIP_DEVICE;

    if (argc < 2) {
            printf("Usage: %s <command> [data]\n"
            "Commands: getport | setport <n> | setdir <n> | setbit <n> | clearbit <n>\n", argv[0]);
            exit(-1);
        }

   char *cmd = argv[1];
   if (strcmp(cmd, "setport") == 0)
   {
     bRequest = SETPORT;
     if (argc < 3) { printf ("Missing argument\n"); exit(-1); }
     wValue = atoi(argv[2]);
   }
   else if (strcmp(cmd, "getport") == 0)
   {
     bRequest = GETPORT;
     bmRequestType = USB_TYPE_VENDOR | USB_ENDPOINT_IN;
     wLength = 8;
   }
   else if (strcmp(cmd, "setdir") == 0)
   {
     bRequest = SETDIR;
     if (argc < 3) { printf ("Missing argument\n"); exit(-1); }
     wValue = atoi(argv[2]);
   }
   else if (strcmp(cmd, "setbit") == 0)
   {
     bRequest = SETBIT;
     if (argc < 3) { printf ("Missing argument\n"); exit(-1); }
     wValue = atoi(argv[2]);
   }
   else if (strcmp(cmd, "clearbit") == 0)
   {
     bRequest = CLEARBIT;
     if (argc < 3) { printf ("Missing argument\n"); exit(-1); }
     wValue = atoi(argv[2]);
   }
   else
   {
       printf("No such command\n");
       exit(-1);
   }

    usb_init();

    usb_find_busses();
    usb_find_devices();

    udev = NULL;
    dev_found = FALSE;
    for (bus = usb_get_busses(); bus && !dev_found; bus = bus->next) {
        for (dev = bus->devices; dev && !dev_found; dev = dev->next) {
            if ((dev->descriptor.idVendor == 0x04D8) && (dev->descriptor.idProduct == 0x0001)) {
                dev_found = TRUE;
                udev = usb_open(dev);
            }
        }
    }
    if (!dev_found) {
        printf("No matching device found...\n");
        exit(-1);
    }

    if (usb_claim_interface(udev, 0) < 0)
    {
        printf("Unable to claim interface\n");
        exit(-1);
    }

    if (udev)
    {
       ret = usb_control_msg(udev, bmRequestType, bRequest, wValue, wIndex, buffer, wLength, 100);
    }

        if (ret < 0) {
            printf("Unable to send vendor specific request, ret = %d...\n", ret);
        } else
        {
         printf("OK: buffer length = %d\n", ret);
         if ( ret > 0 )
         {
           printf("Buffer: %02hhx %02hhx\n", buffer[0], buffer[1]);
         }
        }

  usb_close(udev);
}
